from ctypes import *
from locale import currency
from xdxsmi.api.node import Node
import re
import string
import time
import subprocess
import sys

# buffer size
NVML_DEVICE_NAME_BUFFER_SIZE                 = 64
NVML_DEVICE_PCI_BUS_ID_BUFFER_V2_SIZE        = 16
NVML_DEVICE_PCI_BUS_ID_BUFFER_SIZE           = 32

class _PrintableStructure(Structure):
    """
    Abstract class that produces nicer __str__ output than ctypes.Structure.
    e.g. instead of:
      >>> print str(obj)
      <class_name object at 0x7fdf82fef9e0>
    this class will print
      class_name(field_name: formatted_value, field_name: formatted_value)

    _fmt_ dictionary of <str _field_ name> -> <str format>
    e.g. class that has _field_ 'hex_value', c_uint could be formatted with
      _fmt_ = {"hex_value" : "%08X"}
    to produce nicer output.
    Default fomratting string for all fields can be set with key "<default>" like:
      _fmt_ = {"<default>" : "%d MHz"} # e.g all values are numbers in MHz.
    If not set it's assumed to be just "%s"

    Exact format of returned str from this class is subject to change in the future.
    """
    _fmt_ = {}
    def __str__(self):
        result = []
        for x in self._fields_:
            key = x[0]
            value = getattr(self, key)
            fmt = "%s"
            if key in self._fmt_:
                fmt = self._fmt_[key]
            elif "<default>" in self._fmt_:
                fmt = self._fmt_["<default>"]
            result.append(("%s: " + fmt) % (key, value))
        return self.__class__.__name__ + "(" + ','.join(result) + ")"

class nvmlPciInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('domain', c_uint64),
        ('bus', c_uint64),
        ('device', c_uint64),
        ('bdfid', c_uint64),
        ('func', c_uint64),
        ('pci_dbdf', c_char * 64),
        ('busId', c_char * NVML_DEVICE_PCI_BUS_ID_BUFFER_SIZE),
    ]
    _fmt_ = {
            'domain'         : "0x%08X",
            'bus'            : "0x%02X",
            'device'         : "0x%02X",
            'bdfid'          : "0x%08X",
            }

## Device structures
class struct_c_xdx_device_t(Structure):
    pass # opaque handle
c_xdx_device_t = POINTER(struct_c_xdx_device_t)

#clock info structure
class xdxmlclockInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('xdx_gddr_phys_freq', c_int),
        ('xdx_gfx_freq', c_int),
        ('xdx_acpu_freq', c_int),
        ('xdx_cmcu_freq', c_int),
        ('xdx_ndf_freq', c_int),
    ]
    _fmt_ = {
            'xdx_gddr_phys_freq'      : "%d",
            'xdx_gfx_freq'            : "%d",
            'xdx_acpu_freq'           : "%d",
            'xdx_cmcu_freq'           : "%d",
            'xdx_ndf_freq'            : "%d",
            }

#power info structure
class xdxmlpowerInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('volt', c_int),
        ('current', c_int),
        ('pwr', c_int),
        ('power_cap', c_int),
        ('power1_cap_max', c_int),
        ('power1_cap_min', c_int),
    ]
    _fmt_ = {
            'volt'      : "%d",
            'current'            : "%d",
            'pwr'           : "%d",
            }

#temper info structure
class xdxmltemperInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('ctrl_mode', c_int),
        ('temp', c_int),
        ('temp_highest', c_int),
        ('temp_min', c_int),
        ('temp_max', c_int),
        ('temp_crit', c_int),
        ('gddr_temp', c_int),
    ]

#temper info structure
class xdxmltemperSet_t(_PrintableStructure):
    _fields_ = [
        ('ctrl_mode', c_char* 5),
        ('temp_min', c_char* 5),
        ('temp_max', c_char* 5),
        ('temp_crit', c_char* 5),
    ]

## C Type mappings ##
## Enums
XDXML_FEATURE_DISABLED    = 0
XDXML_FEATURE_ENABLED     = 1

class xdxmlutilInfo_t(_PrintableStructure):
    _fields_ = [
        ('gpu',  c_uint),
        ('tdm',  c_uint),
        ('geom', c_uint),
        ('thrd',   c_uint),
        ('cdm',  c_uint),
    ]

class xdxmldrvInfo_t(_PrintableStructure):
    _fields_ = [
        ('DDK', c_char* 10),
        ('KMD', c_char* 10),
        ('UMD', c_char* 10),
    ]

class xdxmlmemInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('fb_free', c_ulonglong),
        ('fb_total', c_ulonglong),
        ('fb_used', c_ulonglong),
        ('sys_free', c_ulonglong),
        ('sys_total', c_ulonglong),
        ('sys_used', c_ulonglong),
    ]

#temper info structure
class xdxmlqueryInfo_t(_PrintableStructure):
    _fields_ = [
        ('display_status', c_uint),
        # Moved to the new busId location below
        ('xdx_gddr_phys_freq', c_int),
        ('xdx_gfx_freq', c_int),
        ('xdx_acpu_freq', c_int),
        ('xdx_cmcu_freq', c_int),
        ('xdx_ndf_freq', c_int),

        ('name', c_char* NVML_DEVICE_NAME_BUFFER_SIZE),
        ('uuid', c_char* 10),
        ('vbios_version', c_char* 20),
        ('serial', c_char* 10),
        ('minor_number', c_int),
        ('fan_speed', c_int),
        ('primary_id', c_char* 10),
        ('primary_core_id', c_char* 10),
        ('asci_type', c_char* 10),
        ('core', c_int),
        ('mmbar_type', c_char* 20),
        ('fw_ver', c_char* 10),
        ('gddr_count', c_int),
        ('gddr_speed', c_ulong),
        ('gddr_size', c_int),

        #pcie error
        ('err_corr', c_char* 10),
        ('err_fatal', c_char* 10),
        ('non_fatal', c_char* 10),

        ('ctrl_mode', c_int),
        ('temp', c_int),
        ('temp_highest', c_int),
        ('temp_min', c_int),
        ('temp_max', c_int),
        ('temp_crit', c_int),

        ('curr_pcie_link_width', c_uint),
        ('max_pcie_link_width', c_uint),
        ('pcie_speed', c_uint),
        ('pcie_max_speed', c_uint),
        ('curr_link_gen', c_uint),
        ('max_link_gen', c_uint),

        #pcie event
        ('pcie_event', c_int),
        ('perst_assert', c_int),
        ('perst_de_assert', c_int),
        ('link_clk_stable', c_int),
        ('link_down', c_int),
        ('perst_assert_acpu_ack', c_int),
        ('perst_de_assert_acpu_ack', c_int),

        #pcie_info
        ('domain', c_uint64),
        ('bus', c_uint64),
        ('device', c_uint64),
        ('bdfid', c_uint64),
        ('func', c_uint64),
        ('pci_dbdf', c_char * 64),

        #memory info
        ('fb_free', c_ulonglong),
        ('fb_total', c_ulonglong),
        ('fb_used', c_ulonglong),
        ('sys_free', c_ulonglong),
        ('sys_total', c_ulonglong),
        ('sys_used', c_ulonglong),

        #GPU info wcj   
        ('bValid',c_uint64),     
        ('ui64GpuStatActive',c_uint64),
        ('ui64GpuStatBlocked',c_uint64),
        ('ui64GpuStatIdle',c_uint64),
        ('ui64GpuStatCumulative',c_uint64),

        ('aaui64DMOSStatActive_0',c_uint64 ),
        ('aaui64DMOSStatActive_1',c_uint64 ),
        ('aaui64DMOSStatActive_2',c_uint64 ),
        ('aaui64DMOSStatActive_3',c_uint64 ),
        ('aaui64DMOSStatActive_4',c_uint64 ),
        ('aaui64DMOSStatActive_5',c_uint64 ),
        ('aaui64DMOSStatActive_6',c_uint64 ),
        ('aaui64DMOSStatActive_7',c_uint64 ),
        ('aaui64DMOSStatActive_8',c_uint64 ),

        ('aaui64DMOSStatBlocked_0',c_uint64),
        ('aaui64DMOSStatBlocked_1',c_uint64),
        ('aaui64DMOSStatBlocked_2',c_uint64),
        ('aaui64DMOSStatBlocked_3',c_uint64),
        ('aaui64DMOSStatBlocked_4',c_uint64),
        ('aaui64DMOSStatBlocked_5',c_uint64),
        ('aaui64DMOSStatBlocked_6',c_uint64),
        ('aaui64DMOSStatBlocked_7',c_uint64),
        ('aaui64DMOSStatBlocked_8',c_uint64),


        ('aaui64DMOSStatIdle_0',c_uint64),
        ('aaui64DMOSStatIdle_1',c_uint64),
        ('aaui64DMOSStatIdle_2',c_uint64),
        ('aaui64DMOSStatIdle_3',c_uint64),
        ('aaui64DMOSStatIdle_4',c_uint64),
        ('aaui64DMOSStatIdle_5',c_uint64),
        ('aaui64DMOSStatIdle_6',c_uint64),
        ('aaui64DMOSStatIdle_7',c_uint64),
        ('aaui64DMOSStatIdle_8',c_uint64),

        ('aaui64DMOSStatCumulative_0',c_uint64),
        ('aaui64DMOSStatCumulative_1',c_uint64),
        ('aaui64DMOSStatCumulative_2',c_uint64),
        ('aaui64DMOSStatCumulative_3',c_uint64),
        ('aaui64DMOSStatCumulative_4',c_uint64),
        ('aaui64DMOSStatCumulative_5',c_uint64),
        ('aaui64DMOSStatCumulative_6',c_uint64),
        ('aaui64DMOSStatCumulative_7',c_uint64),
        ('aaui64DMOSStatCumulative_8',c_uint64),

        ('ui64TimeStamp',c_uint64),

        ('pronum',c_uint64),

        ('process_1',c_uint64),             
        ('pro1LMA',c_uint64),
        ('pro1Total',c_uint64),

        ('process_2',c_uint64),        
        ('pro2LMA',c_uint64),
        ('pro2Total',c_uint64),

        ('process_3',c_uint64),        
        ('pro3LMA',c_uint64),
        ('pro3Total',c_uint64),

        ('process_4',c_uint64),        
        ('pro4LMA',c_uint64),
        ('pro4Total',c_uint64),

        ('process_5',c_uint64),        
        ('pro5LMA',c_uint64),
        ('pro5Total',c_uint64),

        ('pro1NAME',c_char* 30),   
        ('pro2NAME',c_char* 30),
        ('pro3NAME',c_char* 30),
        ('pro4NAME',c_char* 30),
        ('pro5NAME',c_char* 30),
		('device_id', c_uint64),
        ('volt', c_int), 
        ('support_fan_info', c_int), 
        ('current', c_float), 
        ('pwr', c_float), 
        ('chip_model', c_char* 64), 
        
    ]


def load_xdxml_library():
    '''
    Load the library if it isn't loaded already
    '''
    global xdxml_lib
    etc_node = Node(f"/etc")
    curr_os = etc_node['issue'].split(' ', 1)[0]
    arch = subprocess.check_output('arch', shell = True).decode('utf-8').split('\n',1)[0]#when use check_output the value of arch is 'x86_64\n' but 'x86_64',thus,needed to slipt '\n'
    try:
        if arch == "x86_64":     
            if curr_os == "Ubuntu":
                xdxml_lib = CDLL("/opt/xdxgpu/lib/x86_64-linux-gnu/libxdxgpu-ml.so")
                # xdxml_lib = CDLL("/home/li.li/workspace/8_29rtm/drivers/tools/xdxsmi/xdxml_lib/out/libxdxgpu-ml.so.1.3.0")
            else:
                xdxml_lib = CDLL("/usr/lib/x86_64-linux-gnu/xdxgpu/libxdxgpu-ml.so")

            
        elif arch == "aarch64":
            if curr_os == "Ubuntu":
                xdxml_lib = CDLL("/opt/xdxgpu/lib/aarch64-linux-gnu/libxdxgpu-ml.so")
            else:
                xdxml_lib = CDLL("/usr/lib/aarch64-linux-gnu/xdxgpu/libxdxgpu-ml.so")

    except Exception as e:
        print("Failed to load xdxml library:", str(e))
        exit()


## Function access ##
_xdxmlGetFunctionPointer_cache = dict() # function pointers are cached to prevent unnecessary libLoadLock locking
def _xdxmlGetFunctionPointer(name):
    global xdxml_lib

    if name in _xdxmlGetFunctionPointer_cache:
        return _xdxmlGetFunctionPointer_cache[name]

    _xdxmlGetFunctionPointer_cache[name] = getattr(xdxml_lib, name)
    return _xdxmlGetFunctionPointer_cache[name]

def xdxml_init():
    #print("zihan_debug:enter xdxml_init")
    load_xdxml_library()

    #
    # Initialize the library
    #
    fn = _xdxmlGetFunctionPointer("xdxml_init")
    ret = fn()
    return None

def xdxml_out_of_band_init():
    load_xdxml_library()
    #
    # Initialize the library
    #
    fn = _xdxmlGetFunctionPointer("xdxml_out_of_band_init")
    ret = fn()
    return None

## Device get functions
def xdxml_device_get_count():
    c_count = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_count")
    ret = fn(byref(c_count))
    return c_count.value

def xdxml_device_get_handle_by_index(index):
    c_index = c_uint(index)
    device = c_xdx_device_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_handle_by_index")
    ret = fn(c_index, byref(device))
    return device

# a map function that handle to card index, assist detect_xdxgpus2() in smi.py
def xdxml_device_get_card_index_by_handle(handle):
    c_index = c_uint()
    device = handle
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_card_index_by_handle")
    ret = fn(device, byref(c_index))

    return c_index.value

def xdxml_device_get_name(handle):
    c_name = create_string_buffer(NVML_DEVICE_NAME_BUFFER_SIZE)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_name")
    ret = fn(handle, c_name, c_uint(NVML_DEVICE_NAME_BUFFER_SIZE))
    return c_name.value

def xdxml_device_get_pci_info(handle):
    c_info = nvmlPciInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pci_info")
    ret = fn(handle, byref(c_info))
    #_nvmlCheckReturn(ret)
    return c_info

def xdxml_device_get_clock(handle):
    clock_t = xdxmlclockInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_clock")
    ret = fn(handle, byref(clock_t))
    return clock_t  

def xdxml_device_get_curr_pcie_link_width(handle):
    curr_link_width = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_curr_pcie_link_width")
    ret = fn(handle, byref(curr_link_width))
    return curr_link_width.value

def xdxml_device_get_max_pcie_link_width(handle):
    max_link_width = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_max_pcie_link_width")
    ret = fn(handle, byref(max_link_width))
    return max_link_width.value  

def xdxml_device_get_pcie_speed(handle):
    curr_link_speed = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pcie_speed")
    ret = fn(handle, byref(curr_link_speed))
    return curr_link_speed.value

def xdxml_device_get_pcie_link_max_speed(handle):
    max_link_speed = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pcie_link_max_speed")
    ret = fn(handle, byref(max_link_speed))
    return max_link_speed.value

def xdxml_device_get_curr_pcie_link_generation(handle):
    curr_link_gen = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_curr_pcie_link_generation")
    ret = fn(handle, byref(curr_link_gen))
    return curr_link_gen.value

def xdxml_device_get_max_pcie_link_generation(handle):
    max_link_gen = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_max_pcie_link_generation")
    ret = fn(handle, byref(max_link_gen))
    return max_link_gen.value 

def xdxml_device_get_fan_speed(handle):
    fan_speed = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_fan_speed")
    ret = fn(handle, byref(fan_speed))
    return fan_speed.value

def xdxml_device_get_min_max_fan_speed(handle):
    min = c_int()
    max = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_min_max_fan_speed")
    ret = fn(handle, byref(min), byref(max))
    fan_dict = dict(min_fan_speed = min.value, max_fan_speed = max.value)
    return fan_dict

def xdxml_device_get_fan_pwm(handle):
    fan_pwm = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_fan_pwm")
    ret = fn(handle, byref(fan_pwm))
    return fan_pwm.value

def xdxml_device_get_pwm_min(handle):
    fan_pwm_min = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pwm_min")
    ret = fn(handle, byref(fan_pwm_min))
    return fan_pwm_min.value

def xdxml_device_get_fan_ctrl_mode(handle):
    fan_ctrl_mode = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_fan_ctrl_mode")
    ret = fn(handle, byref(fan_ctrl_mode))
    return fan_ctrl_mode.value

def xdxml_device_get_pwm_max(handle):
    fan_pwm_max = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pwm_max")
    ret = fn(handle, byref(fan_pwm_max))
    return fan_pwm_max.value

def xdxml_device_get_power_info(handle):
    power_t = xdxmlpowerInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_power_info")
    ret = fn(handle, byref(power_t))
    return power_t

def xdxml_device_get_temper_info(handle):
    temper_t = xdxmltemperInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_temper_info")
    ret = fn(handle, byref(temper_t))
    return temper_t

def xdxml_device_get_display_status(handle):
    c_mode = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_display_status")
    ret = fn(handle, byref(c_mode))
    return c_mode.value

def xdxml_device_get_mem_info(handle):
    memory_t = xdxmlmemInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_mem_info")
    ret = fn(handle, byref(memory_t))
    return memory_t

def xdxml_device_get_driver_version(handle):
    driver_ver_t = xdxmldrvInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_driver_version")
    ret = fn(handle, byref(driver_ver_t))
    return driver_ver_t

def xdxml_device_get_util_info(handle):
    util_t = xdxmlutilInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_util_info")
    ret = fn(handle, byref(util_t))
    return util_t

def xdxml_device_query(handle):
    query_t = xdxmlqueryInfo_t()
    fn =  _xdxmlGetFunctionPointer("xdxml_device_query")
    ret = fn(handle, byref(query_t))
    return query_t

def xdxml_device_query_by_out_of_band(handle):
    query_t = xdxmlqueryInfo_t()
    fn =  _xdxmlGetFunctionPointer("xdxml_device_query_by_out_of_band")
    ret = fn(handle, byref(query_t))
    return query_t


#set function
def xdxml_device_set_fan_pwm1(handle, pwm):
    c_data = c_char_p(str(pwm).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_fan_pwm1")
    ret = fn(handle,c_data)

def xdxml_device_set_fan_ctrl_mode(handle, fan_ctrl_mode):
    c_data = c_char_p(str(fan_ctrl_mode).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_fan_ctrl_mode")
    ret = fn(handle,c_data)

def xdxml_device_set_fan1_speed(handle, fan):
    c_data = c_char_p(str(fan).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_fan1_speed")
    ret = fn(handle,c_data)

def xdxml_device_set_power_warn(handle, cap):
    c_data = c_char_p(str(cap).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_power_warn")
    ret = fn(handle,c_data)

def xdxml_device_set_power_volt(handle, input):
    c_data = c_char_p(str(input).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_power_volt")
    ret = fn(handle,c_data)

def xdxml_device_set_gfx_clock(handle, gfx_clock):
    c_data = c_char_p(str(gfx_clock).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_gfx_clock")
    ret = fn(handle,c_data)

def xdxml_device_set_temper_info(handle,temper,tem_type):        
    temper_set = xdxmltemperSet_t()
    temp = str(temper).encode()
    #print("temp is " + temp.decode("utf-8"))  
    if tem_type == "temp1_crit":
        type = c_uint(3)
        temper_set.temp_crit = temp
        #print("PYTHON zihan_debug "+ str(temper_set.temp_crit))
    if tem_type == "temp1_min":
        type = c_uint(1)
        temper_set.temp_min = temp
        #print("temp_min is "+ temper_set.temp_min.decode("utf-8"))
    if tem_type == "temp1_max":
        type = c_uint(2)
        temper_set.temp_max = temp
    if tem_type == "temp1_ctrl_mode":
        type = c_uint(0)
        temper_set.ctrl_mode = temp
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_temper_info")
    ret = fn(handle, byref(temper_set), type)

def xdxml_device_get_support_fan_info(handle):
    support_fan_info = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_support_fan_info")
    ret = fn(handle, byref(support_fan_info))
    return support_fan_info.value

def xdxml_device_get_device_model(handle):
    c_device_model_name = create_string_buffer(NVML_DEVICE_NAME_BUFFER_SIZE)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_device_model")
    ret = fn(handle, c_device_model_name)
    return c_device_model_name.value

def xdxml_device_get_asic_type(handle):
    c_asic_type = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_asic_type")
    ret = fn(handle, byref(c_asic_type))
    return c_asic_type.value
