#! /usr/bin/python3
"""
usage:
    xdxsmi [<command>] [<args>...] [options]
    
commands:
   clock          Control and query clock information of GPUs
   temp           Control and query temperature information of GPUs
   power          Control and query power information of GPUs
   fan            Control and query fan information of GPUs
   list           List all GPUs with their pci info
   query          Query GPUs' attributes
   stats          show statistic information about GPUs
   out            Show gpu info by out of band manager way

options:
    -f, --filename=file  redirect query output to a specified file
    -l, --loop=sec       set look up interval, whose unit is second [default: -1]
    -h, --help           print this help message
    -i, --index=val      target gpu to read/write attributes, start from 0
    --version


See 'xdxsmi <command> -h' for more information on a specific command.
Run 'xdxsmi' without any options or commands will print a general gpu status.
"""

from typing import Dict
from sys import exit
import time
import sys

try:
    from xdxsmi.api.smi import XdxSmi
    from xdxsmi.api.xdxgpu import XdxGpu
    from xdxsmi.api.out_of_band_gpu_info import XdxGpuInfo
    from xdxsmi.api.third_party_docopt import docopt
except ModuleNotFoundError:
    import sys
    from os.path import abspath, join, dirname
    sys.path.insert(0, join(abspath(dirname(dirname(dirname((__file__)))))))
    from xdxsmi.api.smi import XdxSmi
    from xdxsmi.api.xdxgpu import XdxGpu
    from xdxsmi.api.out_of_band_gpu_info import XdxGpuInfo
    from xdxsmi.api.third_party_docopt import docopt


def show_xdxgpu_status_table_header(xdx_smi: XdxSmi, args: Dict):
    smi = xdx_smi.smi_ver
    ver = xdx_smi.vers
    xdxgpu_status_table_header = \
f"""\
+-----------------------------------------------------------------------------------------+
| XDXSMI {smi:16}               KMD Version {ver[2]:9}             UMD Version {ver[3]} |
+----------------------------------+-------------------------------+----------------------+
| GPU     Name        Persistence-M| Bus-Id                 Disp.A | Volatile Uncorr. ECC |
| Fan     Temp  Perf  Pwr:Usage/Cap|                  Memory-Usage |             GPU-Util |
|                                  |                               |                      |
|==================================+===============================+======================|"""

    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write(xdxgpu_status_table_header)
        fo.close()
    else:
        print(xdxgpu_status_table_header)


def show_xdxgpu_status_table_body(xdxgpu: XdxGpu, args: Dict):
    idx = xdxgpu.user_index
    name = xdxgpu.name
    bsid = xdxgpu.pci_bdf
    dsp = xdxgpu.dsp_mode
    dsp_t = xdxgpu.dsp_type
    if xdxgpu.fan_spd=="0":        
        fan = 'N/A'
    else:
        fan = xdxgpu.fan_spd
    fan = str(fan) + "RPM"
    temp = xdxgpu.temp
    fb_used = xdxgpu.fb_used
    fb_total = xdxgpu.fb_total

    xdxgpu_status_table_body = \
f"""\
| {idx}       {name:6}{'...':5} {'N/A':13}|{bsid:14} {''}{dsp_t:4}{''}{dsp:10}| {' ':16} {'N/A'} |
| {fan:7} {temp:6}{'N/A':5} {'N/A':12} | {' ':11} {fb_used:4} MB / {fb_total:4} MB | {' ':16} {'N/A'} |
|                                  |                               |                      |
+----------------------------------+-------------------------------+----------------------+"""

    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write('\n' + xdxgpu_status_table_body)
        fo.close()
    else:
        print(xdxgpu_status_table_body)

def show_xdxgpu_status_table_process_header(xdx_smi: XdxSmi, args: Dict):
    xdxgpu_status_table_process_header = \
f"""\
+-----------------------------------------------------------------------------------------+
| Processes: [Wait to finish]                                                             |
|  GPU   GI   CI        PID   Type   Process name                              GPU Memory |
|        ID   ID                                                               Usage      |
|=========================================================================================|"""
    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write('\n'+ xdxgpu_status_table_process_header)
        fo.close()
    else:
        print(xdxgpu_status_table_process_header)

def show_xdxgpu_status_table_process_body(xdxgpu: XdxGpu, args: Dict):
    
    process_number = xdxgpu.pronum + 1
    default_body = f"""|   {xdxgpu.user_index: >2}{' ': <3}{'N/A': <5}{'N/A': <5}{'': <4}{'N/A': <8}{'N/A': <3}{'': <3}{'N/A': <11}{' ': <30}{'N/A': >8}{' ': <4}|"""
    xdxgpu_status_table_process_body = default_body if process_number == 1 else ''

    for i in range(1, process_number):
        gpu_id = xdxgpu.user_index if xdxgpu.user_index != None else 'N/A'
        pid = getattr(xdxgpu, f"pro{i}PID") if getattr(xdxgpu, f"pro{i}PID") != None else 'N/A'
        type = 'N/A'
        p_name = getattr(xdxgpu,f"pro{i}NAME") if getattr(xdxgpu,f"pro{i}NAME") != None else 'N/A'
        g_usage = getattr(xdxgpu,f"pro{i}Total") if getattr(xdxgpu,f"pro{i}Total") != None else 'N/A'

        line = \
f"""|   {gpu_id: >2}{' ': <3}{'N/A': <5}{'N/A': <5}{'': <4}{pid: <8}{type: <3}{'': <3}{p_name: <11}{' ': <30}{g_usage: >8}MiB{' ': <1}|"""
        
        if(i != xdxgpu.pronum):
            line += '\n'
        xdxgpu_status_table_process_body += line

    # option scheme: seperate processes by gpu number, if so, do not use table foot
    # xdxgpu_status_table_process_body += "\n+-----------------------------------------------------------------------------------------+"
    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write('\n' + xdxgpu_status_table_process_body + '\n')
        fo.close()
    else:
        print(xdxgpu_status_table_process_body)


def show_xdxgpu_status_table_foot(xdx_smi: XdxSmi, args: Dict):
    xdxgpu_status_table_foot = "+-----------------------------------------------------------------------------------------+\n"

    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write(xdxgpu_status_table_foot)
        fo.close()
    else:
        print(xdxgpu_status_table_foot)

# before split only supports 1 gpu
# def show_xdxgpu_status_table_foot(xdxgpu: XdxGpu, args: Dict):
#     xdxgpu_status_table_foot = \
# f"""\
# +-----------------------------------------------------------------------------------------+
# | Processes: [Wait to finish]                                                             |
# |  GPU   GI   CI        PID   Type   Process name                              GPU Memory |
# |        ID   ID                                                               Usage      |
# |=========================================================================================|\n"""

#     for i in range(1, xdxgpu.pronum + 1):
#         gpu_id = xdxgpu.user_index if xdxgpu.user_index != None else 'N/A'
#         pid = getattr(xdxgpu, f"pro{i}PID") if getattr(xdxgpu, f"pro{i}PID") != None else 'N/A'
#         type = 'N/A'
#         p_name = getattr(xdxgpu,f"pro{i}NAME") if getattr(xdxgpu,f"pro{i}NAME") != None else 'N/A'
#         g_usage = getattr(xdxgpu,f"pro{i}Total") if getattr(xdxgpu,f"pro{i}Total") != None else 'N/A'

#         line = \
# f"""|   {gpu_id: >2}{' ': <3}{'N/A': <5}{'N/A': <5}{'': <4}{pid: <8}{type: <3}{'': <3}{p_name: <11}{' ': <30}{g_usage: >8}MiB{' ': <1}|\n"""
#         xdxgpu_status_table_foot += line

#     xdxgpu_status_table_foot += "+-----------------------------------------------------------------------------------------+\n"

#     if(args['--filename'] != None):
#         fo = open(args['--filename'], "a")
#         fo.write('\n' + xdxgpu_status_table_foot + '\n')
#         fo.close()
#     else:
#         print(xdxgpu_status_table_foot)


if __name__ == '__main__':
    args = docopt(__doc__, version="1.4", options_first=True)

    # e.g. xdxsmi -l=1 clock get
    if(args['--loop'] != None and args['--loop'] != '-1'):
        args['<args>'] += ['--loop=' + args['--loop']]

    # e.g. xdxsmi -f=a.txt clock get
    if(args['--filename'] != None):
        args['<args>'] += ['--filename=' + args['--filename']]

    if(args['--index'] != None):
        args['<args>'] += ['--index=' + args['--index']]

    if args['<command>'] == 'out':
        import xdxsmi_out_of_band
        # out command only supports '-h' and '--help' 
        if not(args['--filename'] == None and args['--help'] == False 
               and args['--loop'] == '-1' and args['--version'] == False 
               and (args['<args>'] == [] or args['<args>'] == ['-h'] or args['<args>'] == ['--help'])):
            print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
            exit()
        argv = [args['<command>']] + args['<args>']
        args = docopt(xdxsmi_out_of_band.__doc__, argv=argv)

        xdx_smi = XdxSmi(True)
        xdxsmi_out_of_band.handle(xdx_smi, args)
    else:
        xdx_smi =  XdxSmi()
        # we move the '-f' '-l' before command to the tail of the processing logic
        # so args["<args>"] may not be None.
        if not args["<command>"]:
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1     

            # catch invalid value of '--index'
            index = -1
            num_xdxgpus = len(xdx_smi.xdxgpus)
            try:
                if(args['--index'] == None):
                    index = -1                    # default situation
                elif(args['--index'] != None and int(args['--index']) >= 0
                    and int(args['--index']) < num_xdxgpus):
                    index = int(args['--index'])  # assign index
                else:
                    raise ValueError("No devices were found")
            except:
                print("No devices were found")
                exit()
        
            try:
                while True:
                    target_gpus = xdx_smi.xdxgpus
                    if(index >= 0):
                        target_gpus = [target_gpus[index]]

                    show_xdxgpu_status_table_header(xdx_smi, args)
                    for xdxgpu in target_gpus:
                        show_xdxgpu_status_table_body(xdxgpu, args)
                    # split table foot to three parts can be convenient for multi gpus
                    show_xdxgpu_status_table_process_header(xdx_smi, args)
                    for xdxgpu in target_gpus:
                        show_xdxgpu_status_table_process_body(xdxgpu, args)
                    show_xdxgpu_status_table_foot(xdx_smi, args)
                    
                    
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)                    
                    xdx_smi =  XdxSmi()
            except KeyboardInterrupt:
                exit()
        argv = [args['<command>']] + args['<args>']
        
        # handle options
        if args['<command>'] == 'clock':
            import xdxsmi_clock
            args = docopt(xdxsmi_clock.__doc__, argv=argv)

            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            xdxsmi_clock.handle(xdx_smi, args)

            # if the value of '--loop' is valid, loop regularly
            # otherwise, display only once.
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1   
        
            try:
                # loop_interval = float(args["--loop"])
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)

                    xdx_smi =  XdxSmi()
                    xdxsmi_clock.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit()   
        elif args['<command>'] == 'power':
            import xdxsmi_power
            args = docopt(xdxsmi_power.__doc__, argv=argv)

            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            
            xdxsmi_power.handle(xdx_smi, args)

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 

            try:
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
                    xdx_smi =  XdxSmi()
                    xdxsmi_power.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit()   
        elif args['<command>'] == 'temp':
            import xdxsmi_temp
            args = docopt(xdxsmi_temp.__doc__, argv=argv)
            
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()

            xdxsmi_temp.handle(xdx_smi, args)

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 
            
            try:
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
                    xdx_smi =  XdxSmi()
                    xdxsmi_temp.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit() 

        elif args['<command>'] == 'util':
            import xdxsmi_utility
            args = docopt(xdxsmi_utility.__doc__, argv=argv)

            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            xdxsmi_utility.handle(xdx_smi, args)
        elif args['<command>'] == 'fan':
            import xdxsmi_fan
            args = docopt(xdxsmi_fan.__doc__, argv=argv)
            
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            
            xdxsmi_fan.handle(xdx_smi, args)

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 
            
            try:
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
                    xdx_smi =  XdxSmi()
                    xdxsmi_fan.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit()     
        elif args['<command>'] == 'list':
            import xdxsmi_list
            if not(args['--filename'] == None and args['--help'] == False 
                and args['--loop'] == '-1' and args['--version'] == False 
                and (args['<args>'] == [] or args['<args>'] == ['-h'] or args['<args>'] == ['--help'])):
                print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                exit()
            
            # list command only supports '-h' and '--help'
            args = docopt(xdxsmi_list.__doc__, argv=argv)
            xdxsmi_list.handle(xdx_smi, args)
        elif args['<command>'] == 'set':
            import xdxsmi_set
            args = docopt(xdxsmi_set.__doc__, argv=argv)
            xdxsmi_set.handle(xdx_smi, args)
        elif args['<command>'] == 'query':
            import xdxsmi_query
            args = docopt(xdxsmi_query.__doc__, argv=argv)
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            
            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 

            try:
                while True:
                    xdx_smi =  XdxSmi()
                    xdxsmi_query.handle(xdx_smi, args)

                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
            except KeyboardInterrupt:
                exit()        

        elif args['<command>'] == 'pmon':
            import xdxsmi_pmon
            args = docopt(xdxsmi_pmon.__doc__, argv=argv)
            xdxsmi_pmon.handle(xdx_smi, args)
        elif args['<command>'] == 'stats':
            import xdxsmi_stats
            args = docopt(xdxsmi_stats.__doc__, argv=argv)
            xdxsmi_stats.handle(xdx_smi, args)
        else:
            # same as the nvidia-smi prompt message 
            exit("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
        
