"""
usage:
   xdxsmi power [options] get [<ip>...]
   xdxsmi power [options] set pwr_warn <mw>
   xdxsmi power [options] set volt <mv>

options:
   -h, --help            print this help message
   -f, --filename=file   redirect query output to a specified file
   -l, --loop=sec        set look up interval, whose unit is second [default: -1]
   -i, --index=val       target gpu to read/write attributes, start from 0

ip:
   volt                  voltage of GPU, whose unite is V  
   current               current of GPU, whose unit is A
   pwr                   average power of GPU, whose unit is W
"""

from typing import Dict
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *
from xdxsmi.cli.render import Render

from get_set_attrs import setAttrs, getAttrs
from xdxsmi.api.smi import ASIC_TYPE

#Dictionary matching abbreviated and real attribute names 
power_map = {
   "gpu":   "xdx_dbdf",
   "name":  "xdx_product_name",
   "volt":  "in0_input",
   "curr":  "curr1_input",
   "pwr":   "power1_average",
   "pwr_max":  "power1_cap_max",
   "pwr_min":  "power1_cap_min",
   "pwr_warn": "power1_cap"
}
power_unit_map = {
   "volt": "V",
   "current": "A",
   "pwr": "W"
}

def handle(xdx_smi: XdxSmi, args: Dict):
   ##Collect target gpus
   #target_gpus = [xdx_smi.handle[0]]
   index = -1
   num_xdxgpus = len(xdx_smi.xdxgpus)
   try:
      if(args['--index'] == None):
            index = -1                    # default situation
      elif(args['--index'] != None and int(args['--index']) >= 0
            and int(args['--index']) < num_xdxgpus):
            index = int(args['--index'])  # assign index
      else:
            raise ValueError("No devices were found")
   except:
      print("No devices were found")
      exit()
    
   target_gpus = xdx_smi.xdxgpus
   if(index >= 0):
      target_gpus = [xdx_smi.xdxgpus[index]]
   
   #Set values of attributes 'pwr_warn'
   if args["set"] and args["pwr_warn"] and args["<mw>"]:
      try:
         newVal = str(int(args["<mw>"]))
         for gpu in target_gpus:
            xdev = gpu.handle
            asic_type = xdxml_device_get_asic_type(xdev)
            if asic_type == ASIC_TYPE.CHIP_PANGU_B0.value:
               print("Warning: this card type does not support power relation command")
            else:
               xdxml_device_set_power_warn(xdev, newVal)

      except:
         element_ip = next(iter(args["<mw>"]))
         exit(f"'xdxsmi power set pwr_warn {element_ip}' is not a xdxsmi power command."\
               "See 'xdxsmi power --help'.")

   #Set values of attributes 'volt'
   elif args["set"] and args["volt"] and args["<mv>"]:
      #760 ~ 880 mv
      try:
         for gpu in target_gpus:
            xdev = gpu.user_index
            asic_type = xdxml_device_get_asic_type(xdev)
            if asic_type == ASIC_TYPE.CHIP_PANGU_B0.value:
               print("Warning: this card type does not support power relation command")
            else:
               newVal_int = int(args["<mv>"])
               if newVal_int < 760 or newVal_int > 880:
                  print ("Warning: You can only set voltage between 760 and 880 mv\n")
               newVal = str(newVal_int)
               xdxml_device_set_power_volt(xdev, newVal)

      except:
         element_ip = next(iter(args["<mv>"]))
         exit(f"'xdxsmi power set volt {element_ip}' is not a xdxsmi power command."\
               "See 'xdxsmi power --help'.")

   #Get values of target attributes   
   elif args["get"]:
      target_ip = ["volt", "current", "pwr"]
      # li: filter the same element between target_ip and args["<ip"]
      if args["<ip>"]:
         target_ip = [ip for ip in target_ip if ip in args["<ip>"]]

      # li: judge the input target_ip is legal
      if len(target_ip) == 0:
         element_ip = next(iter(args["<ip>"]))
         exit(f"'xdxsmi power get {element_ip}' is not a xdxsmi power command."\
               "See 'xdxsmi power --help'.")

      data = []
      # i = 0
      header = ["index", "gpu", "name"] + target_ip
      #print(header)
      data.insert(0, header)
      for gpu in target_gpus:
         xdev = gpu.handle
         asic_type = xdxml_device_get_asic_type(xdev)
         if asic_type == ASIC_TYPE.CHIP_PANGU_B0.value:
            print("Warning: this card type does not support power info")
         else:
            powerInfo = xdxml_device_get_power_info(xdev)

         row = [
               str(gpu.user_index),
               xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
               xdxml_device_get_name(xdev).decode("utf8")]
         for ip in target_ip:
            if asic_type == ASIC_TYPE.CHIP_PANGU_B0.value:
               row.append("N/A")
            else:
               row.append(str(round(getattr(powerInfo, ip)/1e3, 2)) + power_unit_map[ip])
         data.append(row)
         # i += 1
      Render.render_table(data, args)
    
   #Stop program if the input command is unvalid
   else:
      exit("%r is not a xdxsmi power command."\
             "See 'xdxsmi power --help'." % args['<command>'])

if __name__ == "__main__":
   print(docopt(__doc__))